# First of all, a simple "unknown" procedure
proc which cmd {
    foreach n "/bin /sbin /usr/bin /usr/sbin /usr/local/bin" {
        if [file executable $n/$cmd] {return $n/$cmd}
    }
    error "$cmd: no such command"
}

proc unknown args {
    set cmd [lindex $args 0]
    if ![catch {which $cmd} res] {
	return [eval exec $res [lrange $args 1 end]]
    }
    error $res
}

# Replace the external command "cat"
if [catch {which cat}] {
    proc cat {args} {
	foreach n $args {
	    set F [open $n]
	    while {[gets $F string]>=0} {puts $string}
	    close $F
	}
    }
}

# Replace "free"
if [catch {which free}] {
    proc free {} {
	set F [open /proc/meminfo]
	gets $F string; gets $F string2
	close $F
	return "$string\n$string2"
    }
}

# Replace "grep"
if [catch {which grep}] {
    proc grep {expr args} {
	foreach n $args {
	    set F [open $n]
	    while {[gets $F string]>=0} {
		if [regexp $expr $string] {puts $string}
	    }
	    close $F
	}
    }
}


# Replace "ls"
if [catch {which ls}] {
    proc ls {args} {
	set ret ""
	foreach n $args {
	    foreach f [glob -nocomplain $n] {
		file lstat $f st
		append ret [format "%s %4o %2i  %5i %5i  %7i %s" \
			[string index $st(type) 0] [expr $st(mode)%010000] \
			$st(nlink) $st(uid) $st(gid) $st(size) $f]
		if ![catch "file readlink $f" link] {
		    append ret " -> $link"
		}
		append ret "\n"
	    }
	}
	return $ret
    }
}

# Replace the commands "cp", "rm", "mv"
if [catch {which cp}] {proc cp {args} {eval file copy $args}}
if [catch {which mv}] {proc mv {args} {eval file rename $args}}
if [catch {which rm}] {proc rm {args} {eval file delete $args}}

# Fake system call
proc sys_getpid {} {return [pid]}

# Simple ps (even if /bin/ps is there)
proc ps {} {
    foreach n [glob /proc/\[1-9\]*] {
	lappend list [file tail $n]
    }
    set list [lsort -int $list]
    foreach n $list {
	if ![catch {set F [open /var/run/$n]}] {
	    set name($n) [gets $F]
	    close $F
	}
    }
    puts "  User  PPid  Pid St Size   RSS  Name"
    puts " -------------------------------------"
    foreach n $list {
	set F [open /proc/$n/status]
	while {[gets $F string]>0} {
	    set v([lindex $string 0]) [lindex $string 1]
	}
	close $F
	if [info exists name($n)] {set v(Name:) $name($n)}
	puts -nonewline [format "%5i %5i %5i %s " \
		$v(Uid:) $v(PPid:) $v(Pid:) $v(State:)]
	if ![info exists v(VmSize:)] {
	    set v(VmSize:) [set v(VmRSS:) 0]
	}
	puts [format "%5i %5i %s" $v(VmSize:) $v(VmRSS:) $v(Name:)]
	unset v
    }
}

# Interact on a file
proc oneline_interact file {
    global interact
    if [gets $file string]<0 {
	exit
    }
    append interact($file) "${string}\n"
    if [info complete  $interact($file)] {
	catch {uplevel #0 "eval \$interact($file)"} result
	set interact($file) ""
	if [string length $result] {puts $file $result}
	catch {sys_wait nohang} res
	puts -nonewline $file "% "
	flush $file
    }
}

proc interact file {
    fconfigure $file -blocking 1
    fconfigure $file -buffering line

    #### recreate your std files
    sys_dup $file stdin
    sys_dup $file stdout
    sys_dup $file stderr
   
    puts -nonewline $file "% "; flush $file
    fileevent $file readable "oneline_interact $file"
}

# MP (14-06-99)
# getAllowedHosts  - parse /etc/hosts.allow
proc getAllowedHosts {service} {
	set allowed_ip {}

	set f [open /etc/hosts.allow]
	while {![eof $f]} {
		gets $f line
		if {[regexp $line {^ *\#}]} {
			continue
		}
		set exp "^${service} *: *(.+)"
		if {[regexp -nocase $exp $line dummy ip_list]} {
			set ip_list [string trim $ip_list]
			set ips [split $ip_list " "]
			foreach ip $ips {
				lappend allowed_ip $ip
			}

			unset dummy
			unset ip
		}
	}
	close $f
	return $allowed_ip
}

# Halt and reboot
proc halt {} {
    open /etc/halt w
    exit
}
proc reboot {} {
    open /etc/reboot w
    exit
}

# netconfig
proc netconfig {} {

    set options_file /etc/init.d/options
    set options2_file /etc/network.options

    proc ask {msg def} {
        puts -nonewline "$msg "
        if [string length $def] {
            puts -nonewline "\[$def\]: "
        }
        flush stdout
        set ans [gets stdin]
        if [string length $ans] {
            return $ans
        }
        return $def
    }

    proc check_dquad {ip} {
        set num [scan $ip "%d.%d.%d.%d" x1 x2 x3 x4]
        if {$num != 4 || $x1 < 0 || $x2 < 0 || $x3 < 0 || $x4 < 0 \
            || $x1 > 255 || $x2 > 255 || $x3 > 255 || $x4 > 255} {
            puts "Invalid input. Quitting."
            exit -1
        }
    }

    source $options_file
    if {$options(reconfigure) == 0} {return}
    puts "Running \"netconfig\" to reconfigure the network"
    puts ""
    set F ""
    foreach nic [array names options if:eth*] {
        set fields [scan $options($nic) "%s %s %s %s" ip mask net ptp]
        regsub if: $nic "" interface

        puts "==================================================="
        puts "device $interface - current configuration:"
        puts ""
        puts " IP      = $ip"
        puts " netmask = $mask"
        puts " network = $net"
        puts ""

        set ans [ask "do you want to change this? (y/n)" ""]
        if {$ans != "y"} {
            continue
        }
        puts ""

        set oldip $ip
        set ip [ask "new IP address?" $ip]
        check_dquad $ip
        if {$ip != $oldip} {
            set mask "255.255.255.0"
            scan $ip "%d.%d.%d.%d" x1 x2 x3 x4
            set net "$x1.$x2.$x3.0"
        }
        set mask [ask "new netmask?" $mask]
        check_dquad $mask
        set net [ask "new network address?" $net]
        check_dquad $net
        scan $net "%d.%d.%d.%d" x1 x2 x3 x4
        set bcast "$x1.$x2.$x3.255"
        set bcast [ask "broadcast address?" $bcast]
        check_dquad $bcast

        puts "re-configuring interface $interface"
        ifconfig $interface down
        ifconfig $interface addr $ip
        ifconfig $interface netmask $mask
        ifconfig $interface broadcast $bcast
        ifconfig $interface up
        route add $net netmask $mask dev $interface

        puts "saving configuration data"
        if ![llength $F] {
            # Use options2, if set
            if [file exists $options2_file] {
                    set F [open $options2_file w]
                    puts $F "# Generated by \"netconfig\", do not edit"
            } else {
                    set F [open $options_file a]
                    puts $F "# The following line(s) added by \"netconfig\""
            }
        }
        puts $F "set options($nic) \"$ip $mask $net\""

        puts ""
        set ans [ask "do you want to set a default route through $interface? (y/n)" ""]
        if {$ans != "y"} {
            continue
        }

        scan $net "%d.%d.%d.%d" x1 x2 x3 x4
        set x4 [ask "default gateway is $x1.$x2.$x3.X, with X=?" ""]
        set gateway "$x1.$x2.$x3.$x4"
        check_dquad $gateway

        puts "adding default route"
        route add default gw $gateway dev $interface

        puts "saving configuration data"
        # F is already opened
        puts $F "set options(defaultroute) \"$gateway $interface\""
    }
    if [llength $F] {close $F}

    puts "==================================================="
    set F [open /etc/resolv.conf r]
    gets $F domains; set domains [lrange $domains 1 end]
    gets $F server; set server [lindex $server 1]
    close $F
    puts "your domains are \"$domains\""
    puts "your name server is \"$server\""
    set ans [ask "do you want to change them? (y/n)" ""]
    if {$ans=="y"} {
	set domains [ask "domain list to search for? " $domains ]
	set server [ask "IP address of your new name server? " $server]
	set F [open /etc/resolv.conf w]
	puts $F "search $domains"
	puts $F "nameserver $server"
	close $F
    } 
    puts "==================================================="
	# MP (14-06-99): added config for HTTPD authorization
	set allowed_ip [getAllowedHosts httpd]
	if {$allowed_ip == {}} {
		puts "No restrictions in effect for HTTPD"
	} else {
		puts "HTTPD authorized IP numbers: $allowed_ip"
	}
    set ans [ask "do you want to change this ? (y/n)" ""]
    if {$ans=="y"} {
		set domains [ask "Authorized IP list ? " $allowed_ip]
		set F [open /etc/hosts.allow w]
		puts $F "# /etc/hosts.allow - Generated by netconfig, do not edit."
		puts $F "# Format:"
		puts $F "#   daemon : ip1 ip2 ..."
		puts $F "httpd: $allowed_ip"
		close $F
    }
    puts "==================================================="
    puts "Done."
    puts ""
}

# tail-f and cat-f spawn a process, they use two common procs
proc internal_handler F {
    if [gets $F string]<0 {close $F; exit}
    puts $string
}
proc internal_follow {file name flag} {
    if [sys_fork] return
    if [sys_fork] exit
    newcmdname "$name $file"
    set F [open $file] 
    if $flag {catch {seek $F -400 end}}
    fileevent $F readable "internal_handler $F"
}
proc tail-f {file} {
    internal_follow $file tail-f 1
}
proc cat-f {file} {
    internal_follow $file cat-f 0
}


